@@ -1,5 +1,5 @@
 /*
- * CU sudo version 1.5.3 (based on Root Group sudo version 1.1)
+ * CU sudo version 1.5.6 (based on Root Group sudo version 1.1)
  *
  * This software comes with no waranty whatsoever, use at your own risk.
  *
@@ -36,7 +36,7 @@
  */
 
 #ifndef lint
-static char rcsid[] = "$Id: check.c,v 1.122 1996/11/14 02:37:16 millert Exp $";
+static char rcsid[] = "$Id: check.c,v 1.140 1998/09/11 23:19:53 millert Exp $";
 #endif /* lint */
 
 #include "config.h"
@@ -77,6 +77,10 @@ static char rcsid[] = "$Id: check.c,v 1.122 1996/11/14 02:37:16 millert Exp $";
 #ifdef HAVE_KERB4
 #  include <krb.h>
 #endif /* HAVE_KERB4 */
+#ifdef HAVE_PAM
+#  include <security/pam_appl.h>
+#  include <security/pam_misc.h>
+#endif /* HAVE_PAM */
 #ifdef HAVE_AFS
 #  include <afs/stds.h>
 #  include <afs/kautils.h>
@@ -112,6 +116,9 @@ static void  reminder			__P((void));
 #ifdef HAVE_KERB4
 static int   sudo_krb_validate_user	__P((struct passwd *, char *));
 #endif /* HAVE_KERB4 */
+#ifdef HAVE_PAM
+static void pam_attempt_auth            __P((void));
+#endif /* HAVE_PAM */
 #ifdef HAVE_SKEY
 static char *sudo_skeyprompt		__P((struct skey *, char *));
 #endif /* HAVE_SKEY */
@@ -124,13 +131,19 @@ int   user_is_exempt			__P((void));
  * Globals
  */
 static int   timedir_is_good;
-static char  timestampfile[MAXPATHLEN + 1];
+static char  timestampfile[MAXPATHLEN];
 #ifdef HAVE_SECURID
 union config_record configure;
 #endif /* HAVE_SECURID */
 #ifdef HAVE_SKEY
 struct skey skey;
 #endif
+#ifdef HAVE_PAM
+static struct pam_conv conv = {
+	misc_conv,
+	NULL
+};
+#endif
 #ifdef HAVE_OPIE
 struct opie opie;
 #endif
@@ -164,7 +177,11 @@ void check_user()
 	if (rtn == 2)
 	    reminder();		/* do the reminder if ticket file is new */
 #endif /* NO_MESSAGE */
+#ifdef HAVE_PAM
+	pam_attempt_auth();
+#else  /* !HAVE_PAM */
 	check_passwd();
+#endif /* HAVE_PAM */
     }
 
     update_timestamp();
@@ -440,7 +457,7 @@ static void check_passwd()
     struct SD_CLIENT sd_dat, *sd;		/* SecurID data block */
     register int counter = TRIES_FOR_PASSWORD;
 
-    (void) memset ((VOID *)&sd_dat, 0, sizeof(sd_dat));
+    (void) memset((VOID *)&sd_dat, 0, sizeof(sd_dat));
     sd = &sd_dat;
 
     /* Initialize SecurID. */
@@ -487,6 +504,10 @@ static void check_passwd()
 #if defined(HAVE_KERB4) && defined(USE_GETPASS)
     char kpass[_PASSWD_LEN + 1];
 #endif /* HAVE_KERB4 && USE_GETPASS */
+#ifdef HAVE_AUTHENTICATE
+    char *message;
+    int reenter;
+#endif /* HAVE_AUTHENTICATE */
 
 #ifdef HAVE_SKEY
     (void) memset((VOID *)&skey, 0, sizeof(skey));
@@ -500,40 +521,51 @@ static void check_passwd()
      */
     while (counter > 0) {
 
-#ifdef HAVE_SKEY
+#ifdef HAVE_AUTHENTICATE
+	/* use AIX authenticate() function */
+#  ifdef USE_GETPASS
+	pass = (char *) getpass(prompt);
+#  else
+	pass = tgetpass(prompt, PASSWORD_TIMEOUT * 60, user_name, shost);
+#  endif /* USE_GETPASS */
+	reenter = 1;
+	if (authenticate(user_name, pass, &reenter, &message) == 0)
+	    return;		/* valid password */
+#else
+#  ifdef HAVE_SKEY
 	/* rewrite the prompt if using s/key since the challenge can change */
 	set_perms(PERM_ROOT, 0);
 	prompt = sudo_skeyprompt(&skey, prompt);
 	set_perms(PERM_USER, 0);
-#endif /* HAVE_SKEY */
-#ifdef HAVE_OPIE
+#  endif /* HAVE_SKEY */
+#  ifdef HAVE_OPIE
 	/* rewrite the prompt if using OPIE since the challenge can change */
 	set_perms(PERM_ROOT, 0);
 	prompt = sudo_opieprompt(&opie, prompt);
 	set_perms(PERM_USER, 0);
-#endif /* HAVE_OPIE */
+#  endif /* HAVE_OPIE */
 
 	/* get a password from the user */
-#ifdef USE_GETPASS
-#  ifdef HAVE_KERB4
+#  ifdef USE_GETPASS
+#    ifdef HAVE_KERB4
 	(void) des_read_pw_string(kpass, sizeof(kpass) - 1, prompt, 0);
 	pass = kpass;
-#  else
+#    else
 	pass = (char *) getpass(prompt);
-#  endif /* HAVE_KERB4 */
-#else
+#    endif /* HAVE_KERB4 */
+#  else
 	pass = tgetpass(prompt, PASSWORD_TIMEOUT * 60, user_name, shost);
-#endif /* USE_GETPASS */
+#  endif /* USE_GETPASS */
 
 	/* Exit loop on nil password */
 	if (!pass || *pass == '\0') {
 	    if (counter == TRIES_FOR_PASSWORD)
-		exit(0);
+		exit(1);
 	    else
 		break;
 	}
 
-#ifdef HAVE_SKEY
+#  ifdef HAVE_SKEY
 	/* Only check s/key db if the user exists there */
 	if (skey.keyfile) {
 	    set_perms(PERM_ROOT, 0);
@@ -543,8 +575,8 @@ static void check_passwd()
 	    }
 	    set_perms(PERM_USER, 0);
 	}
-#endif /* HAVE_SKEY */
-#ifdef HAVE_OPIE
+#  endif /* HAVE_SKEY */
+#  ifdef HAVE_OPIE
 	/* Only check OPIE db if the user exists there */
 	if (opie.opie_flags) {
 	    set_perms(PERM_ROOT, 0);
@@ -554,58 +586,58 @@ static void check_passwd()
 	    }
 	    set_perms(PERM_USER, 0);
 	}
-#endif /* HAVE_OPIE */
-#if !defined(HAVE_SKEY) || !defined(SKEY_ONLY)
+#  endif /* HAVE_OPIE */
+#  if !defined(OTP_ONLY) || (!defined(HAVE_SKEY) && !defined(HAVE_OPIE))
 	/*
 	 * If we use shadow passwords with a different crypt(3)
 	 * check that here, else use standard crypt(3).
 	 */
-#  if (SHADOW_TYPE != SPW_NONE) && (SHADOW_TYPE != SPW_BSD)
-#    if (SHADOW_TYPE == SPW_ULTRIX4)
+#    if (SHADOW_TYPE != SPW_NONE) && (SHADOW_TYPE != SPW_BSD)
+#      if (SHADOW_TYPE == SPW_ULTRIX4)
 	if (!strcmp(user_passwd, (char *) crypt16(pass, user_passwd)))
 	    return;		/* if the passwd is correct return() */
-#    endif /* ULTRIX4 */
-#    if (SHADOW_TYPE == SPW_SECUREWARE) && !defined(__alpha)
-#      ifdef HAVE_BIGCRYPT
+#      endif /* ULTRIX4 */
+#      if (SHADOW_TYPE == SPW_SECUREWARE) && !defined(__alpha)
+#        ifdef HAVE_BIGCRYPT
 	if (strcmp(user_passwd, (char *) bigcrypt(pass, user_passwd)) == 0)
 	    return;           /* if the passwd is correct return() */
-#      else
+#        else
 	if (strcmp(user_passwd, crypt(pass, user_passwd)) == 0)
 	    return;           /* if the passwd is correct return() */
-#      endif /* HAVE_BIGCRYPT */
-#    endif /* SECUREWARE && !__alpha */
-#    if (SHADOW_TYPE == SPW_SECUREWARE) && defined(__alpha)
+#        endif /* HAVE_BIGCRYPT */
+#      endif /* SECUREWARE && !__alpha */
+#      if (SHADOW_TYPE == SPW_SECUREWARE) && defined(__alpha)
 	if (crypt_type == AUTH_CRYPT_BIGCRYPT) {
 	    if (!strcmp(user_passwd, bigcrypt(pass, user_passwd)))
 		return;             /* if the passwd is correct return() */
 	} else if (crypt_type == AUTH_CRYPT_CRYPT16) {
 	    if (!strcmp(user_passwd, crypt16(pass, user_passwd)))
 		return;             /* if the passwd is correct return() */
-#ifdef AUTH_CRYPT_OLDCRYPT
+#        ifdef AUTH_CRYPT_OLDCRYPT
 	} else if (crypt_type == AUTH_CRYPT_OLDCRYPT ||
 		   crypt_type == AUTH_CRYPT_C1CRYPT) {
 	    if (!strcmp(user_passwd, crypt(pass, user_passwd)))
 		return;             /* if the passwd is correct return() */
-#endif
+#        endif
 	} else {
 	    (void) fprintf(stderr,
                     "%s: Sorry, I don't know how to deal with crypt type %d.\n",
                     Argv[0], crypt_type);
 	    exit(1);
 	}
-#    endif /* SECUREWARE && __alpha */
-#  endif /* SHADOW_TYPE != SPW_NONE && SHADOW_TYPE != SPW_BSD */
+#      endif /* SECUREWARE && __alpha */
+#    endif /* SHADOW_TYPE != SPW_NONE && SHADOW_TYPE != SPW_BSD */
 
 	/* Normal UN*X password check */
 	if (!strcmp(user_passwd, (char *) crypt(pass, user_passwd)))
 	    return;		/* if the passwd is correct return() */
 
-#  ifdef HAVE_KERB4
+#    ifdef HAVE_KERB4
 	if (user_uid && sudo_krb_validate_user(user_pw_ent, pass) == 0)
 	    return;
-#  endif /* HAVE_KERB4 */
+#    endif /* HAVE_KERB4 */
 
-#  ifdef HAVE_AFS
+#    ifdef HAVE_AFS
 	if (ka_UserAuthenticateGeneral(KA_USERAUTH_VERSION,
                                        user_name,	/* name */
                                        NULL,		/* instance */
@@ -615,16 +647,17 @@ static void check_passwd()
                                        0, 0,		/* spare */
                                        NULL) == 0)	/* reason */
 	    return;
-#  endif /* HAVE_AFS */
-#  ifdef HAVE_DCE
+#    endif /* HAVE_AFS */
+#    ifdef HAVE_DCE
 	/* 
 	 * consult the DCE registry for password validation
 	 * note that dce_pwent trashes pass upon return...
 	 */
 	if (dce_pwent(user_name, pass))
 	    return;
-#  endif /* HAVE_DCE */
-#endif /* !HAVE_SKEY || !SKEY_ONLY */
+#    endif /* HAVE_DCE */
+#  endif /* !OTP_ONLY || (!HAVE_SKEY && !HAVE_OPIE) */
+#endif /* HAVE_AUTHENTICATE */
 
 	--counter;		/* otherwise, try again  */
 #ifdef USE_INSULTS
@@ -662,9 +695,9 @@ static int sudo_krb_validate_user(pw_ent, pass)
     char tkfile[sizeof(_PATH_SUDO_TIMEDIR) + 4 + MAX_UID_T_LEN];
     int k_errno;
 
-    /* Get the local realm */
+    /* Get the local realm, or retrun failure (no krb.conf) */
     if (krb_get_lrealm(realm, 1) != KSUCCESS)
-	(void) fprintf(stderr, "Warning: Unable to get local kerberos realm\n");
+	return(-1);
 
     /*
      * Set the ticket file to be in sudo sudo timedir so we don't
@@ -699,6 +732,53 @@ static int sudo_krb_validate_user(pw_ent, pass)
 }
 #endif /* HAVE_KERB4 */
 
+#ifdef HAVE_PAM
+/********************************************************************
+ *  pam_attempt_auth()
+ *
+ *  Try to authenticate the user using Pluggable Authentication
+ *  Modules (PAM). Added 9/11/98 by Gary J. Calvin
+ */
+static void pam_attempt_auth()
+{
+    pam_handle_t *pamh=NULL;
+    int retval;
+    register int counter = TRIES_FOR_PASSWORD;
+
+    /* printf("PAM Authentication\n"); */
+    retval = pam_start("sudo", user_name, &conv, &pamh);
+    if (retval != PAM_SUCCESS) {
+        pam_end(pamh, retval);
+        exit(1);
+    }
+    while (counter > 0) {
+        retval = pam_authenticate(pamh, 0);
+        if (retval == PAM_SUCCESS) {
+            set_perms(PERM_USER, 0);
+            pam_end(pamh, retval);
+            return;
+        }
+
+        --counter;
+#ifdef USE_INSULTS
+        (void) fprintf(stderr, "%s\n", INSULT);
+#else
+        (void) fprintf(stderr, "%s\n", INCORRECT_PASSWORD);
+#endif /* USE_INSULTS */
+    }
+    set_perms(PERM_USER, 0);
+
+    if (counter > 0) {
+        log_error(PASSWORD_NOT_CORRECT);
+        inform_user(PASSWORD_NOT_CORRECT);
+    } else {
+        log_error(PASSWORDS_NOT_CORRECT);
+        inform_user(PASSWORDS_NOT_CORRECT);
+    }
+    pam_end(pamh, retval);
+    exit(1);
+}
+#endif /* HAVE_PAM */
 
 #ifdef HAVE_SKEY
 /********************************************************************
@@ -811,16 +891,16 @@ static char *sudo_opieprompt(user_opie, p)
     if ((rval = opiechallenge(user_opie, user_name, challenge)) != 0) {
 #ifdef OTP_ONLY
 	(void) fprintf(stderr,
-		       "%s: You do not exist in the s/key database.\n",
+		       "%s: You do not exist in the opie database.\n",
 		       Argv[0]);
 	exit(1);
 #else
-	/* return the original prompt if we cannot get s/key info */
+	/* return the original prompt if we cannot get opie info */
 	return(orig_prompt);
 #endif /* OTP_ONLY */
     }
 
-    /* get space for new prompt with embedded s/key challenge */
+    /* get space for new prompt with embedded opie challenge */
     if (new_prompt == NULL) {
 	/* allocate space for new prompt */
 	np_size = op_len + strlen(challenge) + 7;
@@ -868,7 +948,7 @@ static void reminder()
     (void) fprintf(stderr, "\n%s\n%s\n\n%s\n%s\n\n",
 #else
     (void) fprintf(stderr, "\n%s\n%s\n%s\n%s\n\n%s\n%s\n\n%s\n%s\n\n",
-	"    CU sudo version 1.5.3, based on Root Group sudo version 1.1",
+	"    CU sudo version , based on Root Group sudo version 1.1",
 	"    sudo version 1.1, Copyright (C) 1991 The Root Group, Inc.",
 	"    sudo comes with ABSOLUTELY NO WARRANTY.  This is free software,",
 	"    and you are welcome to redistribute it under certain conditions.",
@@ -878,7 +958,5 @@ static void reminder()
 	"        #1) Respect the privacy of others.",
 	"        #2) Think before you type."
     );
-    
-    (void) fflush(stderr);
 }
 #endif /* NO_MESSAGE */
